/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.common;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URLConnection;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.springframework.util.MimeTypeUtils;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.DeleteOnCloseInputStream;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import lombok.Getter;
import lombok.extern.log4j.Log4j2;

@Log4j2
public class ZipAttachmentReader {

	public static final String TMP_FILE_SUFFIX = ".ozg-cloud.tmp";
	public static final String SOURCE_ZIP_PREFIX = "zip-attachment";
	static final String TARGET_ATTACHMENT_PREFIX = "ozg_fs_attachment_";

	private static final double ZIP_MAX_THRESHOLD = 100;
	private static final int ZIP_MAX_TOTAL_SIZE = 500 * 1024 * 1024;
	private static final int ZIP_MAX_ENTRIES = 100;

	private final File sourceZipFile;
	@Getter
	private final String sourceFileName;

	ZipAttachmentReader() {
		this.sourceZipFile = null;
		this.sourceFileName = null;
	}

	private ZipAttachmentReader(InputStream sourceZipInputStream, String originalFileName) {
		try {
			this.sourceZipFile = saveSourceZipToFile(sourceZipInputStream, SOURCE_ZIP_PREFIX);
			sourceZipInputStream.close();
		} catch (IOException e) {
			throw new TechnicalException("Can not save original ZIP.", e);
		}
		this.sourceFileName = originalFileName;
	}

	private ZipAttachmentReader(File sourceZipInputFile, String originalFileName) {
		this.sourceZipFile = sourceZipInputFile;
		this.sourceFileName = originalFileName;
	}

	public static ZipAttachmentReader from(InputStream sourceZipInputStream, String originalFileName) {
		return new ZipAttachmentReader(sourceZipInputStream, originalFileName);
	}

	public static ZipAttachmentReader from(File sourceZipInputFile, String originalFileName) {
		return new ZipAttachmentReader(sourceZipInputFile, originalFileName);
	}

	protected static File saveSourceZipToFile(InputStream inputStream, String namePrefix) {
		try {
			File tempFile = File.createTempFile(namePrefix, TMP_FILE_SUFFIX);
			Files.copy(inputStream, tempFile.toPath(), StandardCopyOption.REPLACE_EXISTING);
			tempFile.deleteOnExit();
			return tempFile;
		} catch (IOException e) {
			throw new ReadZipException("Cannot save source ZIP file to local storage. Processing interrupted.", e);
		}
	}

	public List<IncomingFile> readContent() throws ReadZipException {
		try (ZipInputStream sourceZipInputStream = new ZipInputStream(new FileInputStream(sourceZipFile))) {
			return readContent(sourceZipInputStream);
		} catch (IOException e) {
			throw new ReadZipException("Cannot save file contained in ZIP. Processing interrupted.", e);
		}
	}

	public void deleteSourceFile() {
		try {
			Files.delete(sourceZipFile.toPath());
		} catch (IOException e) {
			LOG.error("Error deleting source ZIP file.", e);
		}
	}

	List<IncomingFile> readContent(ZipInputStream sourceZipInputStream) throws IOException, ReadZipException {
		ZipEntry nextEntry = entryExists(sourceZipInputStream.getNextEntry());
		List<IncomingFile> extractedFiles = new ArrayList<>();

		AtomicInteger totalExtractedSize = new AtomicInteger();
		int totalZipEntries = 0;

		while (Objects.nonNull(nextEntry)) {

			final ZipEntry currentEntry = nextEntry;

			Optional.of(currentEntry)
					.filter(entry -> !entry.isDirectory())
					.map(entry -> createLocalTempFile())
					.map(localFile -> {
						int size = saveZipFileToLocalFile(currentEntry, sourceZipInputStream, localFile);
						extractedFiles.add(createContentEntry(localFile, currentEntry));
						return size;
					})
					.ifPresent(totalExtractedSize::addAndGet);

			totalZipEntries++;
			checkTotalExtractedSize(totalExtractedSize.get());
			checkTotalZipEntries(totalZipEntries);

			nextEntry = sourceZipInputStream.getNextEntry();
		}

		return extractedFiles;
	}

	File createLocalTempFile() {
		try {
			File localFile = File.createTempFile(TARGET_ATTACHMENT_PREFIX, TMP_FILE_SUFFIX);
			localFile.deleteOnExit();
			return localFile;
		} catch (IOException e) {
			throw new ReadZipException("Could not create tmp file", e);
		}
	}

	void checkTotalExtractedSize(Integer totalExtractedSize) {
		if (totalExtractedSize > ZIP_MAX_TOTAL_SIZE) {
			throw new ReadZipException("Total size of uncompressed zip file is to high (" + totalExtractedSize + "> " + ZIP_MAX_TOTAL_SIZE + ")");
		}
	}

	private void checkTotalZipEntries(Integer totalZipEntries) {
		if (totalZipEntries > ZIP_MAX_ENTRIES) {
			throw new ReadZipException("Total entries in zip file exceeded (" + totalZipEntries + "> " + ZIP_MAX_ENTRIES + ")");
		}
	}

	private ZipEntry entryExists(ZipEntry entry) {
		if (Objects.isNull(entry)) {
			throw new ReadZipException("Zip archive either invalid or empty.");
		}
		return entry;
	}

	int saveZipFileToLocalFile(ZipEntry zipEntry, InputStream inputStream, File localFile) {
		try {

			try (FileOutputStream out = new FileOutputStream(localFile)) {
				int totalSizeEntry = 0;

				int readBytes = -1;
				byte[] buffer = new byte[2048];
				while ((readBytes = inputStream.read(buffer)) > 0) { // Compliant
					out.write(buffer, 0, readBytes);
					totalSizeEntry += readBytes;

					double compressionRatio = (double) totalSizeEntry / zipEntry.getCompressedSize();
					if (compressionRatio > ZIP_MAX_THRESHOLD) {
						throw new ReadZipException(
								"Ratio between compressed and uncompressed data is highly suspicious (" + compressionRatio
										+ "), looks like a Zip Bomb Attack");
					}
				}

				return totalSizeEntry;
			}
		} catch (IOException e) {
			throw new ReadZipException("Cannot save file contained in ZIP. Processing interrupted.", e);
		}
	}

	IncomingFile createContentEntry(File file, ZipEntry zipEntry) {
		return IncomingFile.builder()
				.name(zipEntry.getName())
				.size(zipEntry.getSize())
				.contentType(getContentType(zipEntry.getName()))
				.file(file)
				.build();
	}

	@Deprecated
	public InputStream getSourceZipAsStream() {
		try {
			return new DeleteOnCloseInputStream(sourceZipFile);
		} catch (FileNotFoundException e) {
			throw new TechnicalException("Original ZIP was deleted", e);
		}
	}

	public File getSourceZip() {
		return sourceZipFile;
	}

	public long getSourceFileSize() {
		try {
			return Files.size(sourceZipFile.toPath());
		} catch (IOException e) {
			throw new TechnicalException("Cannot get size of source ZIP.", e);
		}
	}

	String getContentType(String name) {
		Objects.requireNonNull(name);
		return Objects.requireNonNullElse(URLConnection.guessContentTypeFromName(name), MimeTypeUtils.APPLICATION_OCTET_STREAM_VALUE);
	}
}
